import collections.abc
from typing import Union

import bson
import h5py
from ._vendor import mongomock
from ._vendor.mongomock import store as mongomock_store
from ._vendor.mongomock import thread as mongomock_thread
import numpy as np

from . import stores

__all__ = "ServerStore", "DatabaseStore", "CollectionStore"


class ServerStore(stores.ServerStore):
    def __init__(self, filename: str, mode="a"):
        self._databases = h5py.File(filename, mode)

    def __getitem__(self, db_name) -> "DatabaseStore":
        db: h5py.Group = self._databases.require_group(db_name)
        return DatabaseStore(db)

    def __contains__(self, db_name):
        return self[db_name].is_created

    def list_created_database_names(self):
        return [name for name in self._databases if self[name].is_created]

    def close(self):
        self._databases.close()


class DatabaseStore(mongomock_store.DatabaseStore):
    """Object holding the data for a database (many collections)."""

    def __init__(self, group: h5py.Group):
        self._group = group
        self._collections = {}

    def __getitem__(self, col_name) -> "CollectionStore":
        try:
            return self._collections[col_name]
        except KeyError:
            coll = CollectionStore(self._group.require_group(col_name))
            self._collections[col_name] = coll
            return coll

    def __contains__(self, col_name) -> bool:
        if col_name not in self._collections:
            return False

        return self[col_name].is_created

    def list_created_collection_names(self):
        return [name for name in self._collections if self[name].is_created]

    def create_collection(self, name):
        col = self[name]
        col.create()
        return col

    def rename(self, name, new_name):
        existing = self._group.get(new_name)
        if existing is not None:
            del self._group[new_name]

        self._group.move(name, new_name)
        self._collections[new_name] = self._collections.pop(name)

    @property
    def is_created(self):
        return any(self[coll_name].is_created for coll_name in self._collections)


class CollectionStore(mongomock_store.CollectionStore):
    """Object holding the data for a collection."""

    DOCUMENTS = "documents"
    INDEXES = "indexes"
    IS_FORCE_CREATED = "is_force_created"
    TTL_INDEXES = "ttl_indexes"

    EMPTY_UTF16 = "".encode("utf16")

    def __init__(self, group: h5py.Group):
        self._group = group

        try:
            group.create_dataset(self.IS_FORCE_CREATED, data=False)
        except ValueError:
            pass

        try:
            doc_group = group[self.DOCUMENTS]
        except KeyError:
            doc_group = group.create_group(self.DOCUMENTS, track_order=True)
        self._documents = GroupDict(doc_group)

        self.indexes = IndexDict(group.require_group(self.INDEXES))
        self._ttl_indexes = GroupDict(group.require_group(self.TTL_INDEXES))

        self._rwlock = mongomock_thread.RWLock()

    @property
    def _is_force_created(self) -> bool:
        return self._group[self.IS_FORCE_CREATED][()]

    @_is_force_created.setter
    def _is_force_created(self, new_val: bool):
        self._group[self.IS_FORCE_CREATED][()] = new_val

    def create(self):
        self._is_force_created = True

    @property
    def is_created(self):
        return self._documents or self.indexes or self._is_force_created

    def drop(self):
        self._documents.clear()
        self.indexes.clear()
        self._ttl_indexes.clear()
        self._is_force_created = False

    def create_index(self, index_name: str, index_dict: dict):
        self.indexes[index_name] = index_dict
        if index_dict.get("expireAfterSeconds") is not None:
            self._ttl_indexes[index_name] = index_dict

    def drop_index(self, index_name: str):
        self._remove_expired_documents()

        # The main index object should raise a KeyError, but the
        # TTL indexes have no meaning to the outside.
        del self.indexes[index_name]
        self._ttl_indexes.pop(index_name, None)

    def __contains__(self, key):
        return super().__contains__(self._encode_key(key))

    def __getitem__(self, key: str):
        return super().__getitem__(self._encode_key(key))

    def __setitem__(self, key, val: dict):
        super().__setitem__(self._encode_key(key), val)

    def __delitem__(self, key):
        super().__delitem__(self._encode_key(key))

    def _encode_key(self, key) -> Union[str, bytes]:
        if key == "":
            return self.EMPTY_UTF16

        return str(key)


class GroupDict(collections.abc.MutableMapping):
    """Object that stores MongoDB documents in a HDF5 group

    Storage of MongoDB documents is done as a BSON.encoded uint8 variable length array, using the trick discussed here:
    https://github.com/mila-iqia/fuel/issues/360#issuecomment-237890510
    """

    def __init__(self, group: h5py.Group):
        self._group = group

    @property
    def group(self) -> h5py.Group:
        return self._group

    def __getitem__(self, key: str) -> dict:
        return self._decode_doc(self._group[key])

    def __iter__(self):
        return self._group.__iter__()

    def __len__(self):
        return self._group.__len__()

    def __setitem__(self, key: str, value: dict):
        # Doing this emulates normal python dictionary behaviour
        # i.e. assigning a value to an existing key just overwrites it
        value = self._encode_doc(value)
        try:
            ds = self._group[key]
        except KeyError:
            dt = h5py.vlen_dtype(np.dtype("uint8"))
            ds = self._group.create_dataset(key, shape=(), dtype=dt)

        ds[()] = value

    def __delitem__(self, key: str):
        del self._group[key]

    def _encode_doc(self, doc: dict):
        encoded = bson.encode(doc)
        return np.frombuffer(encoded, dtype="uint8")

    def _decode_doc(self, dataset) -> dict:
        encoded = dataset[()].tobytes()
        return bson.decode(encoded)


class IndexDict(GroupDict):
    """Specialised group dictionary that stores MongoDB index information.

    This is necessary because index key specifications are given as tuples while BSON which is used to store
    the index dictionary converts these to lists.  Here we intercept decoding and convert back to tuples.
    """

    def _decode_doc(self, dataset) -> dict:
        index_dict = super()._decode_doc(dataset)
        # Convert list of index keys to tuples as they should be
        index_dict["key"] = list(map(tuple, index_dict["key"]))
        return index_dict



import os

try:
    from pymongo.errors import PyMongoError
except ImportError:

    class PyMongoError(Exception):
        pass


try:
    from pymongo.errors import OperationFailure
except ImportError:

    class OperationFailure(PyMongoError):
        def __init__(self, message, code=None, details=None):
            super(OperationFailure, self).__init__()
            self._message = message
            self._code = code
            self._details = details

        code = property(lambda self: self._code)
        details = property(lambda self: self._details)

        def __str__(self):
            return self._message


try:
    from pymongo.errors import WriteError
except ImportError:

    class WriteError(OperationFailure):
        pass


try:
    from pymongo.errors import DuplicateKeyError
except ImportError:

    class DuplicateKeyError(WriteError):
        pass


try:
    from pymongo.errors import BulkWriteError
except ImportError:

    class BulkWriteError(OperationFailure):
        def __init__(self, results):
            super(BulkWriteError, self).__init__("batch op errors occurred", 65, results)


try:
    from pymongo.errors import CollectionInvalid
except ImportError:

    class CollectionInvalid(PyMongoError):
        pass


try:
    from pymongo.errors import InvalidName
except ImportError:

    class InvalidName(PyMongoError):
        pass


try:
    from pymongo.errors import InvalidOperation
except ImportError:

    class InvalidOperation(PyMongoError):
        pass


try:
    from pymongo.errors import ConfigurationError
except ImportError:

    class ConfigurationError(PyMongoError):
        pass


try:
    from pymongo.errors import InvalidURI
except ImportError:

    class InvalidURI(ConfigurationError):
        pass


from .helpers import ObjectId, utcnow  # noqa
from .__version__ import __version__


__all__ = [
    "__version__",
    "Database",
    "DuplicateKeyError",
    "Collection",
    "CollectionInvalid",
    "InvalidName",
    "MongoClient",
    "ObjectId",
    "OperationFailure",
    "WriteConcern",
    "ignore_feature",
    "patch",
    "warn_on_feature",
    "SERVER_VERSION",
]

from .collection import Collection
from .database import Database
from .mongo_client import MongoClient
from .patch import patch
from .write_concern import WriteConcern
from .not_implemented import ignore_feature, warn_on_feature

# The version of the server faked by mongomock. Callers may patch it before creating connections to
# update the behavior of mongomock.
# Keep the default version in sync with docker-compose.yml and travis.yml.
SERVER_VERSION = os.getenv("MONGODB", "5.0.5")



try:
    from pymongo.results import BulkWriteResult
    from pymongo.results import DeleteResult
    from pymongo.results import InsertManyResult
    from pymongo.results import InsertOneResult
    from pymongo.results import UpdateResult
except ImportError:

    class _WriteResult(object):
        def __init__(self, acknowledged=True):
            self.__acknowledged = acknowledged

        @property
        def acknowledged(self):
            return self.__acknowledged

    class InsertOneResult(_WriteResult):

        __slots__ = ("__inserted_id", "__acknowledged")

        def __init__(self, inserted_id, acknowledged=True):
            self.__inserted_id = inserted_id
            super(InsertOneResult, self).__init__(acknowledged)

        @property
        def inserted_id(self):
            return self.__inserted_id

    class InsertManyResult(_WriteResult):

        __slots__ = ("__inserted_ids", "__acknowledged")

        def __init__(self, inserted_ids, acknowledged=True):
            self.__inserted_ids = inserted_ids
            super(InsertManyResult, self).__init__(acknowledged)

        @property
        def inserted_ids(self):
            return self.__inserted_ids

    class UpdateResult(_WriteResult):

        __slots__ = ("__raw_result", "__acknowledged")

        def __init__(self, raw_result, acknowledged=True):
            self.__raw_result = raw_result
            super(UpdateResult, self).__init__(acknowledged)

        @property
        def raw_result(self):
            return self.__raw_result

        @property
        def matched_count(self):
            if self.upserted_id is not None:
                return 0
            return self.__raw_result.get("n", 0)

        @property
        def modified_count(self):
            return self.__raw_result.get("nModified")

        @property
        def upserted_id(self):
            return self.__raw_result.get("upserted")

    class DeleteResult(_WriteResult):

        __slots__ = ("__raw_result", "__acknowledged")

        def __init__(self, raw_result, acknowledged=True):
            self.__raw_result = raw_result
            super(DeleteResult, self).__init__(acknowledged)

        @property
        def raw_result(self):
            return self.__raw_result

        @property
        def deleted_count(self):
            return self.__raw_result.get("n", 0)

    class BulkWriteResult(_WriteResult):

        __slots__ = ("__bulk_api_result", "__acknowledged")

        def __init__(self, bulk_api_result, acknowledged):
            self.__bulk_api_result = bulk_api_result
            super(BulkWriteResult, self).__init__(acknowledged)

        @property
        def bulk_api_result(self):
            return self.__bulk_api_result

        @property
        def inserted_count(self):
            return self.__bulk_api_result.get("nInserted")

        @property
        def matched_count(self):
            return self.__bulk_api_result.get("nMatched")

        @property
        def modified_count(self):
            return self.__bulk_api_result.get("nModified")

        @property
        def deleted_count(self):
            return self.__bulk_api_result.get("nRemoved")

        @property
        def upserted_count(self):
            return self.__bulk_api_result.get("nUpserted")

        @property
        def upserted_ids(self):
            if self.__bulk_api_result:
                return dict(
                    (upsert["index"], upsert["_id"]) for upsert in self.bulk_api_result["upserted"]
                )



import collections.abc
import pathlib
import sqlite3
from typing import Iterator, Union, MutableMapping

import bson
import bson.json_util
from ._vendor import mongomock
from ._vendor.mongomock import store as mongomock_store
from ._vendor.mongomock import thread as mongomock_thread

from . import stores

__all__ = "ServerStore", "DatabaseStore", "CollectionStore"


class ServerStore(stores.ServerStore):
    def __init__(self, dirpath: Union[str, pathlib.Path] = ""):
        """
        Create a new server store

        :param dirpath: the directory where the database folders will be stored, defaults to current directory
        """
        self._dirpath = pathlib.Path(dirpath).absolute()
        self._databases = {}

        if self._dirpath.exists():
            if not self._dirpath.is_dir():
                raise ValueError(f"Path exists but is not a directory: {self._dirpath}")
        else:
            self._dirpath.mkdir(parents=True)

    def __getitem__(self, db_name) -> "DatabaseStore":
        if db_name not in self._databases:
            self._databases[db_name] = DatabaseStore(self._dirpath / db_name)

        return self._databases[db_name]

    def __contains__(self, db_name):
        return self[db_name].is_created

    def list_created_database_names(self):
        return [name for name in self._databases if self[name].is_created]

    def close(self):
        pass


class DatabaseStore(mongomock_store.DatabaseStore):
    """Object holding the data for a database (many collections)."""

    def __init__(self, path: pathlib.Path):
        self._dirpath = path.absolute()
        self._collections: MutableMapping[str, CollectionStore] = {}

        if self._dirpath.exists():
            if not self._dirpath.is_dir():
                raise ValueError(f"Path exists but is not a directory: {self._dirpath}")
        else:
            self._dirpath.mkdir(parents=True)

    def __getitem__(self, col_name) -> "CollectionStore":
        if col_name not in self._collections:
            self._collections[col_name] = CollectionStore(self._dirpath / col_name)

        return self._collections[col_name]

    def __contains__(self, col_name) -> bool:
        if col_name not in self._collections:
            return False

        return self[col_name].is_created

    def list_created_collection_names(self):
        return [name for name in self._collections if self[name].is_created]

    def create_collection(self, name):
        col = self[name]
        col.create()
        return col

    def rename(self, name: str, new_name: str):
        coll = self._collections.pop(name)
        coll.rename(new_name)
        self._collections[new_name] = coll

    @property
    def is_created(self):
        return any(self[coll_name].is_created for coll_name in self._collections)


class CollectionStore(mongomock_store.CollectionStore):
    """Object holding the data for a collection."""

    DOCUMENTS = "documents"
    INDEXES = "indexes"

    ID = "_id"
    DOC = "doc"

    _documents = None
    indexes = None

    def __init__(self, filename: pathlib.Path):
        self._path = filename.absolute()
        self._is_force_created = False
        self._ttl_indexes = {}

        self.open()
        self._rwlock = mongomock_thread.RWLock()

    def create(self):
        self._is_force_created = True
        self._path.touch(exist_ok=True)

    @property
    def is_created(self):
        return not self.is_empty or self.indexes or self._is_force_created

    @property
    def documents(self):
        yield from self._documents.values()

    def __len__(self) -> int:
        return len(self._documents)

    def rename(self, new_name: str):
        self.close()
        new_path = self._path.parent / new_name
        self._path.rename(new_path)
        self._path = new_path
        self.open()

    def drop(self):
        self.close()
        try:
            self._path.unlink()
        except FileNotFoundError:
            pass

        self._ttl_indexes.clear()
        self._is_force_created = False

        self.open()

    def open(self):
        self._connection = sqlite3.connect(self._path)
        self._cur = self._connection.cursor()
        self._documents = TableDict(self.DOCUMENTS, self._connection, self._cur)
        self.indexes = IndexDict(self.INDEXES, self._connection, self._cur)

    def close(self):
        self._cur.close()
        self._cur = None
        self._connection.close()
        self._connection = None
        self._documents = None
        self.indexes = None


class TableDict(collections.abc.MutableMapping):
    ID = "_id"
    DOC = "doc"

    def __init__(self, table_name: str, connection: sqlite3.Connection, cursor: sqlite3.Cursor):
        self._table_name = table_name
        self._connection = connection
        self._cur = cursor
        self._init_table()

    def _init_table(self):
        try:
            self._cur.execute(
                f"CREATE TABLE IF NOT EXISTS {self._table_name}("
                f"{self.ID} PRIMARY KEY,"
                f"{self.DOC}"
                f")"
            )
        except sqlite3.OperationalError as exc:
            print(exc)
        else:
            self._connection.commit()

    def __len__(self) -> int:
        res = self._cur.execute(f"select count(*) from {self._table_name}")  # nosec
        counts = res.fetchone()
        return counts[0]

    def __iter__(self):
        cur = self._connection.cursor()
        try:
            for row in cur.execute(f"SELECT {self.ID} from {self._table_name} ORDER BY ROWID"):
                yield row[0]
        finally:
            cur.close()

    def __contains__(self, key) -> bool:
        key = self._encode_key(key)
        res = self._cur.execute(f"SELECT {self.ID} from {self._table_name} where {self.ID}='{key}'")
        return res.fetchone() is not None

    def __getitem__(self, key: str):
        key = self._encode_key(key)
        res = self._cur.execute(
            f"SELECT {self.DOC} from {self._table_name} where {self.ID}='{key}'"
        )
        value = res.fetchone()
        if value is None:
            raise KeyError(key)

        return self._decode_doc(value[0])

    def __setitem__(self, key, val: dict):
        key = self._encode_key(key)
        val = self._encode_doc(val)
        data = (key, val, val)
        with self._connection:
            self._cur.execute(
                f"INSERT INTO {self._table_name} VALUES(?, ?)"  # nosec
                f"ON CONFLICT({self.ID}) DO UPDATE SET {self.DOC}=?",  # nosec
                data,
            )

    def __delitem__(self, key):
        key = self._encode_key(key)
        with self._connection:
            res = self._cur.execute(
                f"DELETE FROM {self._table_name} WHERE {self.ID}='{key}'"  # nosec
            )
        if res.rowcount <= 0:
            raise KeyError(key)

    def values(self) -> Iterator[dict]:
        cur = self._connection.cursor()
        try:
            for row in cur.execute(f"SELECT {self.DOC} from {self._table_name} ORDER BY ROWID"):
                yield self._decode_doc(row[0])
        finally:
            cur.close()

    def _encode_key(self, key) -> str:
        return str(key)

    def _encode_doc(self, doc: dict) -> Union[str, bytes]:
        """Encode the document dictionary for storing in the backend

        We have also tried bson.json_util.dumps(doc) but this adds a significant computational cost.

        :param doc: the document to be encoded
        :return: the encoded version that can be decoded back using _decode_doc()
        """
        return bson.encode(doc)

    def _decode_doc(self, encoded: Union[str, bytes]) -> dict:
        """
        Decode a document dictionary previously encoded using _encode_doc()

        :param encoded: the encoded document
        :return: the decoded document dictionary
        """
        return bson.decode(encoded)


class IndexDict(TableDict):
    """Specialised group dictionary that stores MongoDB index information.

    This is necessary because index key specifications are given as tuples while BSON which is used to store
    the index dictionary converts these to lists.  Here we intercept decoding and convert back to tuples.
    """

    def _decode_doc(self, dataset) -> dict:
        index_dict = super()._decode_doc(dataset)
        # Convert list of index keys to tuples as they should be
        index_dict["key"] = list(map(tuple, index_dict["key"]))
        return index_dict



import datetime
import decimal
from platform import python_version
import re
import uuid

try:
    from bson import decimal128, Regex
    _HAVE_PYMONGO = True
except ImportError:
    _HAVE_PYMONGO = False


class _NO_VALUE(object):
    pass


# we don't use NOTHING because it might be returned from various APIs
NO_VALUE = _NO_VALUE()

_SUPPORTED_BASE_TYPES = (
    float, bool, str, datetime.datetime, type(None), uuid.UUID, int, bytes, type,
    type(re.compile('')),)

if _HAVE_PYMONGO:
    _SUPPORTED_TYPES = _SUPPORTED_BASE_TYPES + (decimal.Decimal, decimal128.Decimal128)
else:
    _SUPPORTED_TYPES = _SUPPORTED_BASE_TYPES

if python_version() < '3.0':
    dict_type = dict
else:
    from collections import abc
    dict_type = abc.Mapping


def diff(a, b, path=None):
    path = _make_path(path)
    if isinstance(a, (list, tuple)) and isinstance(b, (list, tuple)):
        return _diff_sequences(a, b, path)
    if type(a).__name__ == 'SON':
        a = dict(a)
    if type(b).__name__ == 'SON':
        b = dict(b)
    if type(a).__name__ == 'DBRef':
        a = a.as_doc()
    if type(b).__name__ == 'DBRef':
        b = b.as_doc()
    if isinstance(a, dict_type) and isinstance(b, dict_type):
        return _diff_dicts(a, b, path)
    if type(a).__name__ == 'ObjectId':
        a = str(a)
    if type(b).__name__ == 'ObjectId':
        b = str(b)
    if type(a).__name__ == 'Int64':
        a = int(a)
    if type(b).__name__ == 'Int64':
        b = int(b)
    if _HAVE_PYMONGO and isinstance(a, Regex):
        a = a.try_compile()
    if _HAVE_PYMONGO and isinstance(b, Regex):
        b = b.try_compile()
    if isinstance(a, (list, tuple)) or isinstance(b, (list, tuple)) or \
            isinstance(a, dict_type) or isinstance(b, dict_type):
        return [(path[:], a, b)]
    if not isinstance(a, _SUPPORTED_TYPES):
        raise NotImplementedError(
            'Unsupported diff type: {0}'.format(type(a)))  # pragma: no cover
    if not isinstance(b, _SUPPORTED_TYPES):
        raise NotImplementedError(
            'Unsupported diff type: {0}'.format(type(b)))  # pragma: no cover
    if a != b:
        return [(path[:], a, b)]
    return []


def _diff_dicts(a, b, path):
    if not isinstance(a, type(b)):
        return [(path[:], type(a), type(b))]
    returned = []
    for key in set(a) | set(b):
        a_value = a.get(key, NO_VALUE)
        b_value = b.get(key, NO_VALUE)
        path.append(key)
        if a_value is NO_VALUE or b_value is NO_VALUE:
            returned.append((path[:], a_value, b_value))
        else:
            returned.extend(diff(a_value, b_value, path))
        path.pop()
    return returned


def _diff_sequences(a, b, path):
    if len(a) != len(b):
        return [(path[:], a, b)]
    returned = []
    for i, a_i in enumerate(a):
        path.append(i)
        returned.extend(diff(a_i, b[i], path))
        path.pop()
    return returned


def _make_path(path):
    if path is None:
        return []
    return path



"""
Please refer to the documentation provided in the README.md
"""

from . import stores
from . import version
from .version import *
from .stores import *
from .client import *

__all__ = stores.__all__ + version.__all__ + client.__all__ + ('stores', 'version')
